/*
 * Copyright (c) 2000, 2001 Murray Stokely
 * Copyright (c) 2000, 2001 BSDi
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Driver for National Instruments PCI E-Series data acquisition cards.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, January 1999 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *
 */

#include <sys/param.h>  /* Necessary defines */
#include <sys/systm.h>  /* uprintf */ 
#include <sys/conf.h>   /* cdevsw struct */
#include <sys/uio.h>    /* uio struct */
#include <sys/malloc.h>
#include <sys/fcntl.h>		/* So we can get open modes */
#include <machine/bus.h>        /* Necessary for bus_space stuff */
#include <machine/resource.h>   /* .. */
#include <sys/bus.h>            /* .. */
#include <sys/rman.h>           /* .. */
#include <sys/types.h>
#include <sys/pciio.h>          /* For IOCTL _IOWR macros */
#include <sys/mutex.h>
#include <vm/vm.h>      /* vtophys */
#include <vm/pmap.h>    /* vtophys */

#include "niio.h"
#include "ni.h"
#include "nireg.h"


int 
ai_ioctl(struct ni_softc *sc, u_long cmd, caddr_t data, int fflag, struct proc *p)
{
    int16_t status;
    struct ai_channel *chan = (struct ai_channel *)data;
    struct ai_scaninfo *scans = (struct ai_scaninfo *)data;
#ifdef TRYDMA
    DMABufferNode *node;
#endif

    switch (cmd) {
    case NICTL_AI_SCAN_START:
	/*
	  status = DAQ_STC_Windowed_Mode_Read(sc, NI_AI_STATUS1);
	  uprintf("(scan_start) status : 0x%04x\n", status);
	*/
	if (sc->scan_count > 1) {
	    AI_Arming(sc);
	} 
/*
	uprintf("AI_SCAN_START\n"
		"  Scan count : %d\n", sc->scan_count);
	uprintf("<Sending a `%s' signal>\n", sc->scan_count > 1 ? "start1" : "convert");
*/

	AI_Start_The_Acquisition(sc);
	return 0;
    case NICTL_AI_SCAN_SETUP:
	mtx_lock(&sc->lock);
	sc->scan_count      = scans->scans;
	sc->scan_interval   = scans->scan_interval;
	sc->sample_interval = scans->sample_interval;
#ifdef TRYDMA
	sc->use |= AI_DMA;
#endif
	mtx_unlock(&sc->lock);

#ifdef DEBUG
	uprintf("AI_SCAN_SETUP\n"
		"  Scan count : %d\n"
		"  Scan interval : %d\n"
		"  Sample interval : %d\n"
		"  Channels(%d) : \n", sc->scan_count, sc->scan_interval, sc->sample_interval, sc->ai_chancnt);
#endif
	/* Step 1 is done during module load */
	/* Step 2, setup the channels */
	for (status = 0; status < sc->ai_chancnt; status++) {
	    AI_Add_Channel(sc, sc->ai_channels[status]);
	}
	MSC_Clock_Configure(sc); /* Step 3 */
	Clear_FIFO(sc);		 /* Step 4 */
	AI_Reset_All(sc);	 /* Step 5 */
	AI_Board_Personalize(sc); /* Step 6 */
	AI_Initialize_Configuration_Memory_Output(sc); /* Step 7 */
	AI_Board_Environmentalize(sc); /* Step 8 */
	AI_Trigger_Signals(sc);	/* Step 9 */
	if (sc->scan_count > 1) {
	    Number_of_Scans(sc);
	}
	AI_Scan_Start(sc);	/* Step 10 */
	AI_End_of_Scan(sc);	/* Step 11 */
	Convert_Signal(sc);
	Clear_FIFO(sc);		/* Step 12 */
	AI_Interrupt_Enable(sc);
	/*	MSC_IRQ_Configure(sc); *//*enable interrupts! */
	/* Now if interrupts, then we call AI_Interrupt_Enable */

	/* --- DMA setup ------------------ */

	/* Always allocate the dma_buffer list as a continuous block */
#ifdef DEBUG
	printf("SAMPLE_SIZE = %d  sc->scan_count : %d  sc->ai_chancnt: %d\n", SAMPLE_SIZE, sc->scan_count, sc->ai_chancnt);
#endif
#ifdef TRYDMA
	node = sc->dma_buffer_head = 
	    (DMABufferNode *)malloc(sc->ai_chancnt * SAMPLE_SIZE * sc->scan_count, 
				    M_NIDMA, M_ZERO);
	node->size = sc->ai_chancnt * SAMPLE_SIZE * sc->scan_count;
	node->virtual = 
	    (uint32_t)malloc(sc->ai_chancnt * SAMPLE_SIZE * sc->scan_count,
			       M_NIDMA, M_ZERO);
	node->physical = vtophys(node->virtual);
	node->next = 0;
	printf("AI dma setup node->size : %d\n"
	       "dma_buffer_head->size : %d\n", node->size, sc->dma_buffer_head->size);
	mMITE_DMA_Program(sc, 1, 0, 0);
#endif

	return 0;
    case NICTL_AI_RESET:
	AI_Reset_All(sc);
	return 0;
    case NICTL_AI_ADDCH:
	sc->ai_channels[sc->ai_chancnt++] = Encode_Channel(chan);
	return 0;
    case NICTL_AI_ONEPT_SETUP:
	/* XXX Unimplemented */
	return 0;
	break;
    case NICTL_AI_CONVERT:

	/* The steps in this IOCTL exactly follow the examples in
           section 4-11 through 4-13 of [2] */

	Configure_Board(sc, Encode_Channel(chan));     /* Step 2 */
	MSC_Clock_Configure(sc);                       /* Step 3 */
	Clear_FIFO(sc);		                       /* Step 4 */
	AI_Reset_All(sc);	                       /* Step 5 */
	AI_Board_Personalize(sc);                      /* Step 6 */
	AI_Initialize_Configuration_Memory_Output(sc); /* Step 7 */
	AI_Board_Environmentalize(sc);                 /* Step 8 */
	AI_Trigger_Signals(sc);	                       /* Step 9 */
	AI_Scan_Start(sc);	                       /* Step 10 */
	AI_End_of_Scan(sc);	                       /* Step 11 */
	Clear_FIFO(sc);		                       /* Step 12 */
	AI_Interrupt_Enable(sc);
	/*	MSC_IRQ_Configure(sc); */
	AI_Start_The_Acquisition(sc);                  /* Step 13 */
	return 0;
	break;
    case NICTL_AI_SETSCAN:
	sc->scan_count      = scans->scans;
	sc->scan_interval   = scans->scan_interval;
	sc->sample_interval = scans->sample_interval;
	return 0;
	break;
    default:
	return -EINVAL;
    }
}

/*
 * This function turns information in our channel structure into the
 * 32bit value that the hardware expects
 * maybe encode_channel?
 */

uint32_t
Encode_Channel(struct ai_channel *chan) {
    uint16_t lowmem = 0, himem = 0;
    switch (chan->mode) {
    case DIFF:
	himem = NIB_DIFFERENTIAL;
	break;
    case RSE:
	himem = NIB_RSE;
	break;
    case NRSE:
	himem = NIB_NRSE;
	break;
    }

    himem |= chan->num;
    lowmem = chan->gain;
    if (chan->polarity)
	lowmem |= NIB_POLARITY;
    lowmem |= NIB_LAST_CHANNEL;
    return ((himem << 16) | lowmem);
}

void
AI_Arming(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND1, NIB_AI_SC_ARM |
				NIB_AI_SI_ARM | NIB_AI_SI2_ARM | NIB_AI_DIV_ARM);
}

void
Convert_Signal(struct ni_softc *sc) {
    /* If we've only got one sample, then return */
    if (sc->sample_interval < 2)
	return;
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SI2_LOAD_A, sc->sample_interval - 1);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SI2_LOAD_B, sc->sample_interval - 1);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_MODE2, NIB_AI_SI2_RELOAD_MODE);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND1, NIB_AI_SI2_LOAD);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_MODE2, NIB_AI_SI2_INITIAL_LOAD_SOURCE);
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
Number_of_Scans(struct ni_softc *sc) {
    int scans = sc->scan_count - 1;
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);

    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SC_LOAD_A1, (scans >> 16) & 0xFFFF);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SC_LOAD_A2, scans & 0xFFFF);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND1, NIB_AI_SC_LOAD);

    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
AI_Add_Channel(struct ni_softc *sc, uint32_t arg) {
  /*    uprintf("Add channel : 0x%x\n", arg);
    uprintf("Himem : 0x%4x\n", (arg >> 16) & 0xffff);
    uprintf("Lomem : 0x%4x\n", arg & 0xffff);
  */
/*    DAQ_STC_Windowed_Mode_Write(sc, NI_WRITE_STROBE0, 1); */
/*    DAQ_STC_Windowed_Mode_Write(sc, NI_WRITE_STROBE1, 1); */
    board_write(sc, NI_CONFIGURATION_MEMORY_HIGH, (arg >> 16) & 0xffff);
    board_write(sc, NI_CONFIGURATION_MEMORY_LOW, arg & 0xffff);
    return;
}

/* This function follows exactly from the pseudo-code in [1] 2.6.3.1 */

void 
AI_Reset_All(struct ni_softc *sc) {
    /* Begin critical section, disable interrupts */

    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_RESET);
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    /* Disable all Interrupts */
    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ENABLE, 0);
    /*    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_A_ENABLE, 0, 
				       NIB_AI_SC_TC_INTERRUPT_ENABLE |
				       NIB_AI_START1_INTERRUPT_ENABLE |
				       NIB_AI_START2_INTERRUPT_ENABLE |
				       NIB_AI_START_INTERRUPT_ENABLE |
				       NIB_AI_STOP_INTERRUPT_ENABLE |
				       NIB_AI_ERROR_INTERRUPT_ENABLE |
				       NIB_AI_FIFO_INTERRUPT_ENABLE);
    */

    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, 0xFFE0);
    /*
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_A_ACK, 0xFFFF,
				       NIB_AI_SC_TC_ERROR_CONFIRM |
				       NIB_AI_SC_TC_INTERRUPT_ACK |
				       NIB_AI_START1_INTERRUPT_ACK |
				       NIB_AI_START2_INTERRUPT_ACK |
				       NIB_AI_START_INTERRUPT_ACK |
				       NIB_AI_STOP_INTERRUPT_ACK |
				       NIB_AI_ERROR_INTERRUPT_ACK);
    */

    /* Clear software copies of the registers! */
    sc->reg[NI_AI_COMMAND1]          = 0;
    sc->reg[NI_AI_COMMAND2]          = 0;
    sc->reg[NI_AI_MODE1]             = 0;
    sc->reg[NI_AI_MODE2]             = 0;
    sc->reg[NI_AI_MODE3]             = 0;
    sc->reg[NI_AI_OUTPUT_CONTROL]    = 0;
    sc->reg[NI_AI_PERSONAL]          = 0;
    sc->reg[NI_AI_START_STOP_SELECT] = 0;
    sc->reg[NI_AI_TRIGGER_SELECT]    = 0;

    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_MODE1, NIB_AI_RESERVED_ONE |
				NIB_AI_START_STOP | NIB_AI_TRIGGER_ONCE);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_MODE2, 0);
/* From Comedi Driver */
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_MODE3, 0);

    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_TRIGGER_SELECT, 0x0060);

    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}


/* 
 * Configure the board for a single channel conversion
 */

void
Configure_Board(struct ni_softc *sc, uint32_t channel) {
    /* This clears the configuration memory */
    DAQ_STC_Windowed_Mode_Write(sc, NI_WRITE_STROBE0, 1);
    /* This clears the ADC FIFO */
    DAQ_STC_Windowed_Mode_Write(sc, NI_WRITE_STROBE1, 1);
    board_write(sc, NI_CONFIGURATION_MEMORY_HIGH, (channel >> 16) & 0xffff);
    board_write(sc, NI_CONFIGURATION_MEMORY_LOW, channel & 0xffff);
}

/* Select the timebase for the DAQ-STC from [1] 4-10 */

void
MSC_Clock_Configure(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Write(sc, NI_CLOCK_AND_FOUT, 
				NIB_SLOW_INTERNAL_TIMEBASE |
				NIB_SLOW_INTERNAL_TIME_DIVIDE |
				NIB_CLOCK_TO_BOARD |
				NIB_CLOCK_TO_BOARD_DIVIDE);
}

void
Clear_FIFO(struct ni_softc *sc) {
    /* 
     * "Clearing the configuration memory and ADC FIFO require
     * windowed mode writes as well as board writes" [2] 4-9"
     */
    DAQ_STC_Windowed_Mode_Write(sc, NI_WRITE_STROBE1, 1);
    board_write(sc, NI_WRITE_STROBE1, 1);
}

void
Clear_Configuration_Memory(struct ni_softc *sc) {
    /* 
     * "Clearing the configuration memory and ADC FIFO require
     * windowed mode writes as well as board writes" [2] 4-9"
     */
    DAQ_STC_Windowed_Mode_Write(sc, NI_WRITE_STROBE0, 1);
    board_write(sc, NI_WRITE_STROBE0, 1);
}

void
AI_Board_Personalize(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    DAQ_STC_Windowed_Mode_Write(sc, NI_CLOCK_AND_FOUT, NIB_AI_OUTPUT_DIVIDE);
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_AI_PERSONAL, 0xA4A0, 0xFDB0);
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_AI_OUTPUT_CONTROL, 0x032E, 0x033F);
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
AI_Initialize_Configuration_Memory_Output(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND1, NIB_AI_CONVERT_PULSE);
}

void
MSC_IRQ_Configure(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_CONTROL, NIB_INTERRUPT_OUTPUT_POLARITY, NIB_INTERRUPT_OUTPUT_POLARITY);
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_CONTROL, NIB_INTERRUPT_A_ENABLE, NIB_INTERRUPT_A_ENABLE);
}


void
AI_Ack_All(struct ni_softc *sc)
{
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);

    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, 0xFFE0);
    /*
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_A_ACK, 0xFFFF,
				       NIB_AI_SC_TC_ERROR_CONFIRM |
				       NIB_AI_SC_TC_INTERRUPT_ACK |
				       NIB_AI_START1_INTERRUPT_ACK |
				       NIB_AI_START2_INTERRUPT_ACK |
				       NIB_AI_START_INTERRUPT_ACK |
				       NIB_AI_STOP_INTERRUPT_ACK |
   				       NIB_AI_ERROR_INTERRUPT_ACK);
    */

    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

/*   Fixed AI_Interrupt_Enable()  */
void 
AI_Interrupt_Enable(struct ni_softc *sc)
{ 
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    /* 
     * MSC_IRQ_Configure routine.
     */
    /* Route STC interrupt 0 to the MITE or MiniMITE */
    /*   
     * Writing to register Interrupt_Control_Register with address 59.
     *    Interrupt_A_Output_Select <= p->msc_irq_pin[irq_group] (4)  
     * New pattern = 0x0001
     */
    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_CONTROL, 0x0001);
    /*
     * Writing to register Interrupt_A_Enable_Register with address 73.
     *   AI_FIFO_Interrupt_Enable <= p->ai_fifo_interrupt (1)
     *   AI_STOP_Interrupt_Enable <= p->ai_stop_interrupt (0)
     *   AI_START_Interrupt_Enable <= p->ai_start_interrupt (0)
     *   AI_SC_TC_Interrupt_Enable <= p->ai_sc_tc_interrupt (0)
     *   AI_START1_Interrupt_Enable <= p->ai_start1_interrupt (0)
     *   AI_START2_Interrupt_Enable <= p->ai_start2_interrupt (0)
     *   AI_Error_Interrupt_Enable <= p->ai_error_interrupt (1)
     * New pattern = 0x00A0
     */

    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ENABLE, NIB_AI_FIFO_INTERRUPT_ENABLE | NIB_AI_ERROR_INTERRUPT_ENABLE);
    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_B_ENABLE, 0); /* XXX just added this */
    /* | NIB_AI_START_INTERRUPT_ENABLE | NIB_AI_STOP_INTERRUPT_ENABLE);*/
    
    /* 
     * Writing to register Interrupt_Control_Register with address 59.
     *    Interrupt_A_Enable <= p->msc_irq_enable[irq_group] (1)
     * New pattern = 0x0801
     */
    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_CONTROL, 0x0801);
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
    return;
}

/*
void
AI_Interrupt_Enable(struct ni_softc *sc) {
    printf("In AI_Interrupt_Enable!\n");
*/
    /*    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START); */
    /* Temporarily from the Linux Driver */
/*
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_CONTROL, 0x0001, 0x0001);
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_CONTROL, 0, 0x0700);
*/
    /* what the hell is this really writing? */
/*    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_A_ENABLE, 
 */
				       /*
				       NIB_AI_START_INTERRUPT_ENABLE |
				       NIB_AI_START1_INTERRUPT_ENABLE |
				       */
/*				       NIB_AI_FIFO_INTERRUPT_ENABLE |
				       NIB_AI_ERROR_INTERRUPT_ENABLE,
*/				       /*
				       NIB_AI_START_INTERRUPT_ENABLE |
				       NIB_AI_START1_INTERRUPT_ENABLE |
				       */
/*				       NIB_AI_FIFO_INTERRUPT_ENABLE |
				       NIB_AI_ERROR_INTERRUPT_ENABLE);
*/
/*
				       NIB_AI_FIFO_INTERRUPT_ENABLE | 
				       NIB_AI_ERROR_INTERRUPT_ENABLE |
				       NIB_AI_STOP_INTERRUPT_ENABLE);
*/
    /* We do this, like we're supposed to .. */
/*
    DAQ_STC_Windowed_Mode_Write(sc, NI_SECOND_IRQ_B_ENABLE, 0x0800); 
*/
    /* This works */
    /* DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_INTERRUPT_CONTROL, 0x0800, 0x0800); */

    /*    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}
    */

void
AI_Board_Environmentalize(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_AI_MODE2, 0, 
				       NIB_AI_EXTERNAL_MUX_PRESENT);
    DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_AI_OUTPUT_CONTROL, 0x0000,
				       0x00C0);
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
AI_Trigger_Signals(struct ni_softc *sc) {
    printf("in AI_Trigger_Signals!\n");
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_MODE1, NIB_AI_TRIGGER_ONCE);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_TRIGGER_SELECT, NIB_AI_START1_SYNC |
				NIB_AI_START1_EDGE);

    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
AI_Scan_Start(struct ni_softc *sc) {
    printf("AI_Scan_Start!\n");
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_START_STOP_SELECT, 
				NIB_AI_START_EDGE | NIB_AI_START_SYNC);

    if (sc->scan_interval > 0) {
	printf("Setting up scan interval! : %d\n", sc->scan_interval);
	/* Now if we've got some scan intervals set, do stuff here */
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SI_LOAD_A1, 0);
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SI_LOAD_A2, 1);
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND1, NIB_AI_SI_LOAD);
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SI_LOAD_A1, (sc->scan_interval >> 16) & 0xFFFF);
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_SI_LOAD_A2, sc->scan_interval & 0xFFFF);
    }

    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
AI_End_of_Scan(struct ni_softc *sc) {
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_START);
    /* XXX generalize me! */
    DAQ_STC_Windowed_Mode_Write(sc, NI_AI_START_STOP_SELECT, 0x2980);
    DAQ_STC_Windowed_Mode_Write(sc, NI_JOINT_RESET, NIB_AI_CONFIGURATION_END);
}

void
AI_Start_The_Acquisition(struct ni_softc *sc) {
    if (sc->scan_count > 1) {
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND2, NIB_AI_START1_PULSE);
    } else {
	DAQ_STC_Windowed_Mode_Write(sc, NI_AI_COMMAND1, NIB_AI_CONVERT_PULSE);
    }
}
