/*
 * Copyright (c) 2000, 2001 Murray Stokely, BSDi
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Driver for National Instruments Data Aquisition cards.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, January 1999 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *
 * Murray Stokely
 * 19 Nov 2000
 */

#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>  /* uprintf */ 
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>   /* cdevsw struct */
#include <sys/uio.h>    /* uio struct */
#include <sys/malloc.h>
#include <sys/fcntl.h>		/* So we can get open modes */
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>	/* structs, prototypes for pci bus stuff */
#include <sys/rman.h>
#include <sys/types.h>
#include <sys/mutex.h>

#include "niio.h"
#include "ni.h"
#include "nireg.h"

extern int quiet;

/*
 * [1] Section 7.6.1 : Windowed Mode Register Access Example
 */

void 
DAQ_STC_Windowed_Mode_Write1(struct ni_softc *sc, uint16_t address, uint8_t value) {
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_ADDRESS * 2, address);
    bus_space_write_1(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_DATA * 2, value);
    sc->reg[address] = value;
}

void 
DAQ_STC_Windowed_Mode_Write(struct ni_softc *sc, uint16_t address, uint16_t value) {
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_ADDRESS * 2, address);
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_DATA * 2, value);
    sc->reg[address] = value;
}

void
DAQ_STC_Windowed_Mode_Masked_Write(struct ni_softc *sc, uint16_t address, uint16_t value, uint16_t mask) {
    uint16_t copy;
    copy = (sc->reg[address] & ~mask) | (value & mask);
    sc->reg[address] = copy;
/*    uprintf("value = 0x%04x, mask = 0x%04x Writing 0x%04x to 0x%04x\n", value, mask, copy, address); */
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_ADDRESS * 2, address);
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_DATA * 2, copy);
}  

void 
DAQ_STC_Windowed_Mode_Strobe_Write(struct ni_softc *sc, uint16_t address, uint16_t value) {
    uint8_t copy;
    copy = sc->reg[address] | value;
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_ADDRESS * 2, address);
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_DATA * 2, copy);
}

uint8_t
DAQ_STC_Windowed_Mode_Read1(struct ni_softc *sc, uint16_t address) {
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_ADDRESS * 2, address);
    return bus_space_read_1(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_DATA * 2);
}

uint16_t
DAQ_STC_Windowed_Mode_Read(struct ni_softc *sc, uint16_t address) {
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_ADDRESS * 2, address);
    return bus_space_read_2(sc->bar1_bt, sc->bar1_bh, NI_WINDOW_DATA * 2);
}


void
board_write(struct ni_softc *sc, uint16_t address, uint16_t value) {
    bus_space_write_2(sc->bar1_bt, sc->bar1_bh, address, value);
}

void
board_write1(struct ni_softc *sc, uint16_t address, uint8_t value) {
    bus_space_write_1(sc->bar1_bt, sc->bar1_bh, address, value);
}

uint16_t
board_read(struct ni_softc *sc, uint16_t address) {
    return bus_space_read_2(sc->bar1_bt, sc->bar1_bh, address);
}

uint16_t
board_read1(struct ni_softc *sc, uint8_t address) {
    return bus_space_read_1(sc->bar1_bt, sc->bar1_bh, address);
}

/* Access the MITE registers */

void
mite_write2(struct ni_softc *sc, uint16_t address, uint16_t value) {
    bus_space_write_2(sc->bar0_bt, sc->bar0_bh, address, value);
}

void
mite_write4(struct ni_softc *sc, uint16_t address, uint32_t value) {
    printf("mite_write4 writing 0x%04x to 0x%04x\n", value, address);
    bus_space_write_4(sc->bar0_bt, sc->bar0_bh, address, value);
}

uint16_t
mite_read2(struct ni_softc *sc, uint16_t address) {
    return bus_space_read_2(sc->bar0_bt, sc->bar0_bh, address);
}

uint32_t
mite_read4(struct ni_softc *sc, uint16_t address) {
    return bus_space_read_4(sc->bar0_bt, sc->bar0_bh, address);
}
