/*
 * Copyright (c) 2001 Murray Stokely
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This code is derived from the PCI-Eser.zip code from National Instruments.
 * It is repsonsible for talking to the MITE chip to setup DMA transfers.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, January 1999 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *  [4] "MITE DMA Information", Tim Hayles October 19, 2000
 *
 */

#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/systm.h>  /* uprintf */ 
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/proc.h>
#include <sys/malloc.h>
#include <sys/mutex.h>
#include <vm/vm.h>      /* vtophys */
#include <vm/pmap.h>    /* vtophys */

#include "niio.h"
#include "ni.h"
#include "mitereg.h"

/* Free the memory used for a Link Chain list */
void
mMITE_Free_Chain(struct ni_softc *sc)
{
    printf("in mMITE_Free_Chain!\n");
    if (!sc->MITELinkChainHead)
	return;
    free(sc->MITELinkChainHead, M_NIDMA);
    sc->MITELinkChainHead = 0;
}

/*
 * Turn our DMABuffer list into a Link Chain list that the mMITE chip can
 * understand and then program the mMITE.
 */

void
mMITE_DMA_Program(struct ni_softc *sc, short buffers, short channel, short drq)
{
    int counter;
    MITELinkChain *MITELinkChainPtr;
    DMABufferNode *dma_buffer_ptr = sc->dma_buffer_head;
    
    if (sc->MITELinkChainHead)
	mMITE_Free_Chain(sc);
    /* Allocate space for the Link Chain + 1 terminal node */
    MITELinkChainPtr = sc->MITELinkChainHead = (MITELinkChain *)
	malloc((buffers + 1) * sizeof(MITELinkChain), M_NIDMA, M_ZERO);

    /* Copy the dma_buffer to the MITELinkChain */
    for (counter=0; counter < buffers; counter++) {
	MITELinkChainPtr->TCR  = dma_buffer_ptr->size;
	MITELinkChainPtr->MAR  = dma_buffer_ptr->physical;
	MITELinkChainPtr->LKAR = vtobus(MITELinkChainPtr) + sizeof(MITELinkChain);
	MITELinkChainPtr++;
	if (dma_buffer_ptr->next)
	    dma_buffer_ptr++;
    }

    /* Create the last node of the MITELinkChain (TCR = 0). */
    MITELinkChainPtr->TCR  = 0;
    MITELinkChainPtr->MAR  = 0;
    MITELinkChainPtr->LKAR = 0;

    /* [4] A9. */
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_CHANNEL_OPERATION,
		CHOR_DMA_RESET | CHOR_FRESET | CHOR_CLR_DONE);
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_CHANNEL_CONTROL,
		CHCR_BITS | CHCR_SET_DMA_IE | CHCR_SET_DONE_IE | 
		CHCR_DIR_INPUT | DMA_LINK_SHORT_MODE);
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_MEMORY_CONFIGURATION,
		0x00E00600);
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_DEVICE_CONFIGURATION,
		0x00E40241);
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_LINK_CONFIGURATION,
		0x00E00700);
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_LINK_ADDRESS,
		vtobus(sc->MITELinkChainHead));
}

/*
 * Start DMA transfers whenever there is data in the FIFO 
 */

void
mMITE_DMA_Arm(struct ni_softc *sc, int channel) {
    printf("\n------[ DMA ARM ]-----------------\n");
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_CHANNEL_OPERATION, CHOR_START);
}

/*
 * Abort DMA transfer
 */
void
mMITE_DMA_Disarm(struct ni_softc *sc, int channel) {
    printf("\n------[ DMA DISARM ]-----------------\n");
    mite_write4(sc, DMA_BASE_ADDR(channel) + DMA_CHANNEL_OPERATION, CHOR_ABORT);
    printf("About to free MITELinkChain\n");
    mMITE_Free_Chain(sc);
}

/*
 * Free the DMA buffers
 */
void
Free_DMA(struct ni_softc *sc) {
    DMABufferNode *node = sc->dma_buffer_head;
    if (!sc->dma_buffer_head)
	return;
    for (node = sc->dma_buffer_head; ; node++) {
	free((caddr_t)node->virtual, M_NIDMA);
	if(node->next == 0)
	    break;
    }
    free(sc->dma_buffer_head, M_NIDMA);
    sc->dma_buffer_head = 0;
}

/*
 * Read the mMITE registers
 */

void 
mMITE_Status(struct ni_softc *sc, int channel) {
    printf("TCR = %d   MAR = 0x%04x   LKAR = 0x%04x\n", 
	   mite_read4(sc, DMA_BASE_ADDR(channel) + DMA_TRANSFER_COUNT), 
	   mite_read4(sc, DMA_BASE_ADDR(channel) + DMA_MEMORY_ADDRESS), 
	   mite_read4(sc, DMA_BASE_ADDR(channel) + DMA_LINK_ADDRESS));
}
