/*
 * Copyright (c) 2000, 2001 Murray Stokely
 * Copyright (c) 2000, 2001 BSDi
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Driver for National Instruments PCI E-Series data aquisition cards.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, January 1999 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *
 */

/*
 * Gain is the multiplier used on the difference between the input voltage
 * and voltage reference before feeding to the ADC to turn into a binary
 * number.
 */


#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>  /* uprintf */ 
#include <sys/errno.h>
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>   /* cdevsw struct */
#include <sys/uio.h>    /* uio struct */
#include <sys/malloc.h>
#include <sys/fcntl.h>		/* So we can get open modes */
#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>	/* structs, prototypes for pci bus stuff */
#include <sys/rman.h>
/* include <sys/resource.h> */
#include <sys/pciio.h>  /* For IOCTL _IOWR macros */
#include <sys/mutex.h>
#include <sys/time.h>
#include <vm/vm.h>      /* vtophys */
#include <vm/pmap.h>    /* vtophys */

#include <pci/pcivar.h> /* For get_pci macros! */
#include <pci/pcireg.h> /* For PCIR_ defines */


#include "niio.h"
#include "ni.h"
#include "nireg.h"
#include "ni_devices.h"

#define MIN(a,b) (((a) < (b)) ? (a) : (b))

/*
 * The ADC FIFO size is 512 (From [3] A-14) but there's no reason why we can't
 * keep more than that in memory.
 */

#define ADC_FIFO_SIZE 2048
#define DMA_BUF_ENTRIES  20
#define DMA_BUF_EACH     2
#define DMA_BUF_SIZE     (DMA_BUF_ENTRIES * DMA_BUF_EACH)

MALLOC_DEFINE(M_NIFIFO, "NIFIFO", "Software buffer of NI ADC FIFO");
MALLOC_DEFINE(M_NICHAN, "NICHAN", "Software buffer of NI ADC Channels");
MALLOC_DEFINE(M_NIDMA, "NIDMA", "Software buffer of NI DMA stuff");
MALLOC_DEFINE(M_NIMITE, "NIMITE", "DMA buffer for mMITE chip");

/* Function prototypes */
d_open_t      ni_open;
d_close_t     ni_close;
d_read_t      ni_read;
d_write_t     ni_write;
d_ioctl_t     ni_ioctl;

/* Character device entry points */

static struct cdevsw ni_cdevsw = {
    ni_open,
    ni_close,
    ni_read,
    ni_write,
    ni_ioctl,
    nopoll,
    nommap,
    nostrategy,
    "ni",
    37,                   /* reserved for lkms - /usr/src/sys/conf/majors */
    nodump,
    nopsize,
    D_TTY,
/* XXX this is not tested, -1 works but causes a compiler warning. maybe this works? */
    NULL,
/*    -1 */
};

/* vars */

static devclass_t ni_devclass;
static int counter=0;
int quiet = 0;

void
ni_handler(void *xsc) {
    struct ni_softc *sc = xsc;
    uint16_t status;
    uint32_t dmastatus = 0;
    static int overruns = 0;
    uint32_t identify;
    struct timeval tv;
    /* The unit number is not the minor number!  */
    int saved;
    int done  = 0;
    int count = 0;
    static int int_count = 0;

    quiet     = 1;


    /* Check to see if our device has a pending interrupt in the case
       that we might be sharing the interrupt with other PCI devices.
       From [2] 4-57 */

    /* XXX check with gcc -S the output of this function and work on
       ways to minimize it */

    identify = mite_read4(sc, 0x14);
    if (!(identify & NIB_INTERRUPT)) {
	return;
    }

    microtime(&tv);
    if (int_count++ < 100)
	printf("(1st) tv_sec: %ld tv_usec: %ld\n", tv.tv_sec, tv.tv_usec);

    saved     = board_read(sc, NI_WINDOW_ADDRESS);
    status    = DAQ_STC_Windowed_Mode_Read(sc, NI_AI_STATUS1);

    if (!(status & NIB_INTERRUPT_A_ST)) {
	if (counter > 10) {
	    AI_Reset_All(sc);
	} else {
	    printf("Non-group A interrupt? : 0x%04x\n", status);
	    counter++;
	}
    }

    while (!done) {
	if (sc->reg[NI_INTERRUPT_A_ENABLE] & NIB_AI_FIFO_INTERRUPT_ENABLE) {
	    while (status & NIB_AI_FIFO_REQUEST_ST) {
		if (count++ > 512)
		    break;
		printf("FIFO interrupt(%d)! : 0x%04x", count, status);

		/* XXX How should interrupts work with DMA code.  Do
                   we only arm the mMITE with the first FIFO
                   interrupt?  What do we do in future interrupts?
                   Should these even occur? */

		/* Are we doing DMA? */
		if (sc->use & AI_DMA) {
		    mMITE_DMA_Arm(sc, 0);
		} else {
		    sc->fifo[sc->fifocnt++] = board_read(sc, NI_ADC_FIFO_DATA);
		    printf(" data: %d", sc->fifo[sc->fifocnt - 1]);
		    if (sc->fifocnt > ADC_FIFO_SIZE) {
			printf("ni: FIFO buffer overflow (read the data you are generating)\n");
			sc->fifocnt = ADC_FIFO_SIZE;
		    }
		    status = DAQ_STC_Windowed_Mode_Read(sc, NI_AI_STATUS1);
		    printf(" newstatus: 0x%04x\n", status);
		    done = 1;
		    microtime(&tv);
		    printf("(2nd) tv_sec: %ld tv_usec: %ld\n", tv.tv_sec, tv.tv_usec);
		}
	    }
	    if (status & NIB_AI_OVERRUN_ST) {
		printf("ADC overrun error.  Increase your interval!\n");
		if (overruns) {
		    sc->sample_interval *= 10;
		    sc->scan_interval *= 10;
		} else {
		    sc->sample_interval += 10;
		    sc->scan_interval += 10;
		}
		DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, NIB_AI_ERROR_INTERRUPT_ACK);
		done = 1;
	    }
	}
	if (sc->reg[NI_INTERRUPT_A_ENABLE] & NIB_AI_START_INTERRUPT_ENABLE) {
	    if (status & NIB_AI_START_ST) {
		DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, NIB_AI_START_INTERRUPT_ACK);
		done = 1;
	    }
	}
	if (sc->reg[NI_INTERRUPT_A_ENABLE] & NIB_AI_STOP_INTERRUPT_ENABLE) {
	    if (status & NIB_AI_STOP_ST) {
		DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, 
					    NIB_AI_STOP_INTERRUPT_ACK);
		done = 1;
	    }
	}
	if (sc->reg[NI_INTERRUPT_A_ENABLE] & NIB_G0_TC_INTERRUPT_ENABLE) {
	    if (status & NIB_G0_TC_ST) {
		/* Ack it! */
		DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, 
					    NIB_G0_TC_INTERRUPT_ACK);
		printf("G0_TC interrupt!\n");
		done = 1;
	    }
	}
	if (sc->reg[NI_INTERRUPT_A_ENABLE] & NIB_G0_GATE_INTERRUPT_ENABLE) {
	    if (status & NIB_G0_GATE_INTERRUPT_ST) {
		DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, 
					    NIB_G0_GATE_INTERRUPT_ACK);
		printf("G0_GATE interrupt!\n");
		done = 1;
	    }
	}
#ifdef TRYDMA
	dmastatus = mite_read4(sc, DMA_BASE_ADDR(0) + DMA_CHANNEL_STATUS);
	if ( dmastatus & 0xF0000000 ) {
	    printf("--------[ DMA interrupt ]----------\n");
	    if ( dmastatus & 0x02000000) {
		printf("sc->dma_buffer_head->size : %d\n"
		       "sc->dma_buffer_head->physical(0x%x) virt(0x%x)\n",
		       sc->dma_buffer_head->size, 
		       sc->dma_buffer_head->physical,
		       sc->dma_buffer_head->virtual);
		printf("value[0] : %d\n", ((uint16_t *)sc->dma_buffer_head->virtual)[0]);
		printf("DMA DONE\n");
	    }
	    if ( dmastatus & 0x0000F000)
		printf("DMA ERROR\n");
	    /* Clear the DMA interrupt biatch */
	    mite_write4(sc, DMA_BASE_ADDR(0) + DMA_CHANNEL_OPERATION,
			0x00000080);
	}
#endif
	if (!done) {
	    if (int_count < 10) {
		printf("AI_Status_1 : 0x%04x  soft: 0x%04x dmastatus : 0x%04x identify: 0x%04x\n", status, sc->reg[NI_INTERRUPT_A_ENABLE], dmastatus, identify);
		AI_Reset_All(sc);
		mMITE_Status(sc, 0);
	    }
	}	
	done = 1;
    } 
    DAQ_STC_Windowed_Mode_Write(sc, NI_INTERRUPT_A_ACK, 0xFFE0);
    board_write(sc, NI_WINDOW_ADDRESS, saved);
    quiet = 0;
}

/*
 * Open should behave differently for the Digital I/O device depending
 * on which open flags are set.  If this is a read only open, then all
 * DIO lines should be programmed as inputs.  If this is a write-only
 * open, then all DIO lines should be programmed as outputs.  For a read-
 * write open, we should split up the input/outputs somehow (configured with
 * an IOCTL?
 */

int 
ni_open(dev_t kdev, int oflags, int devtype, struct proc *p)
{
    struct ni_softc *sc;
    /* The unit number is not the minor number!  */
    device_t dev = devclass_get_device(ni_devclass, minor(kdev) / 16);
    int err = 0;

    sc = (struct ni_softc *)device_get_softc(dev);

    switch ( minor(kdev) % 16) {
    case DEV_AI:
	/* Only one open allowed for doing analog input */
	mtx_lock(&sc->lock);
	/* XXX What if an interrupt occurs while we own this lock? */
	if (sc->use & AI_BUSY) {
	    err = EBUSY;
	    mtx_unlock(&sc->lock);
	    break;
	}
	sc->scan_count = 0;
	sc->scan_interval = 0;
	sc->sample_interval = 0;
	sc->fifocnt  = 0;
	sc->startcnt = 0;
	sc->use |= AI_BUSY;
	mtx_unlock(&sc->lock);
	break;
    case DEV_DIO:
	mtx_lock(&sc->lock);
	if (sc->use & DIO_BUSY) {
	    err = EBUSY;
	    mtx_unlock(&sc->lock);
	    break;
	}
	sc->dio_flags = oflags;
	sc->use |= DIO_BUSY;
	mtx_unlock(&sc->lock);
	break;
    case DEV_GPCT0:
	mtx_lock(&sc->lock);
	if (sc->use & GPCT0_BUSY) {
	    err = EBUSY;
	    mtx_unlock(&sc->lock);
	    break;
	}
	sc->use |= GPCT0_BUSY;
	mtx_unlock(&sc->lock);
	break;
    case DEV_GPCT1:
	mtx_lock(&sc->lock);
	if (sc->use & GPCT1_BUSY) {
	    err = EBUSY;
	    mtx_unlock(&sc->lock);
	    break;
	}
	sc->use |= GPCT1_BUSY;
	mtx_unlock(&sc->lock);
	break;
    case DEV_STATS:
	uprintf("Stats open!\n");
	break;
    default:
	err = ENODEV;
    }

    return(err);
}

/*
 * Simply "closes" our device that was opened with ni_open.
 */

int 
ni_close(dev_t kdev, int fflag, int devtype, struct proc *p)
{
    struct ni_softc *sc;
    device_t dev = devclass_get_device(ni_devclass, minor(kdev) / 16);
    int err=0;

    sc = (struct ni_softc *)device_get_softc(dev);

    switch (minor(kdev) % 16) {
    case DEV_AI:
	mtx_lock(&sc->lock);
	sc->use &= ~AI_BUSY;
	sc->ai_chancnt = 0;
	sc->use &= ~AI_DMA;
	mtx_unlock(&sc->lock);
#ifdef TRYDMA
	mMITE_DMA_Disarm(sc, 0);
#endif
	break;
    case DEV_DIO:
	sc->dio_flags = 0;
	mtx_lock(&sc->lock);
	sc->use &= ~DIO_BUSY;
	mtx_unlock(&sc->lock);
	break;
    case DEV_GPCT0:
	mtx_lock(&sc->lock);
	sc->use &= ~GPCT0_BUSY;
	mtx_unlock(&sc->lock);
    case DEV_GPCT1:
	mtx_lock(&sc->lock);
	sc->use &= ~GPCT1_BUSY;
	mtx_unlock(&sc->lock);
	break;
    case DEV_STATS:
	break;
    default:
	uprintf("Unknown device!\n");
	return ENODEV;
    }

    return(err);
} 

/*
 * The read function just takes the buf that was saved 
 * via ni_write() and returns it to userland for
 * accessing.
 */

int
ni_read(dev_t kdev, struct uio *uio, int ioflag)
{
    struct ni_softc *sc;
    device_t dev = devclass_get_device(ni_devclass, minor(kdev) / 16);
    int err = 0;
    uint32_t save1, save2;
    int amount;
/*    uint16_t i; */

    sc = (struct ni_softc *)device_get_softc(dev);

    switch (minor(kdev) % 16) {
    case DEV_AI:
	/* Here we should allocate a DMA buffer and instruct the
           hardware to transfer its data.  The process will be put to
           sleep. */


        /* The hardware writes data to the DMA buffer and raises an
	   intterupt when its done. */

	/* The interrupt handler gets the input data, acknowledges the
           interrupt, and awakens the process, which is now able to
           read data. */

	/* We should read in the minimum of the request and the
           available data */
	if (sc->fifocnt > sc->startcnt)
	    amount = MIN( (sc->fifocnt - sc->startcnt) , uio->uio_resid);
	else {
	    uprintf("Can't read, fifocnt: %d startcnt: %d\n",sc->fifocnt, sc->startcnt);
	    break;
	}
	err = uiomove((caddr_t)&sc->fifo[sc->startcnt], amount * 2, uio);

	/*	i = sc->fifo[0];
	uprintf("i : %d\n", i);
	err = uiomove((caddr_t)sc->fifo, 2, uio); */
	sc->startcnt += amount;
	break;
    case DEV_DIO:
	sc->dio = DAQ_STC_Windowed_Mode_Read(sc, NI_DIO_PARALLEL_INPUT);
	if ((err = uiomove((caddr_t)&sc->dio, 1, uio)) != 0) {
	    uprintf("uiomove failed!\n");
	}
	break;
    case DEV_GPCT0:
	/* From [1] 4.6.1.13 */
	uprintf("IN GPCT0 READ!\n");
	DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_G0_COMMAND,
					   0, NIB_G0_SAVE_TRACE);
	DAQ_STC_Windowed_Mode_Masked_Write(sc, NI_G0_COMMAND,
					   NIB_G0_SAVE_TRACE,
					   NIB_G0_SAVE_TRACE);
	save1 = DAQ_STC_Windowed_Mode_Read(sc, NI_G0_HW_SAVE1);
	uprintf("Save1 : %d\n", save1);
	save1 <<= 16;
	save1 += DAQ_STC_Windowed_Mode_Read(sc, NI_G0_HW_SAVE2);
	uprintf("Save1 : %d\n", save1);
	save2 = DAQ_STC_Windowed_Mode_Read(sc, NI_G0_HW_SAVE1);
	save2 <<= 16;
	save2 += DAQ_STC_Windowed_Mode_Read(sc, NI_G0_HW_SAVE2);
	if (save1 != save2) {
	    save1 = DAQ_STC_Windowed_Mode_Read(sc, NI_G0_HW_SAVE1);
	    save1 <<= 16;
	    save1 += DAQ_STC_Windowed_Mode_Read(sc, NI_G0_HW_SAVE2);
	    uprintf("Save1 : %d\n", save1);
	}
	if ((err = uiomove((caddr_t)&save1, 4, uio)) != 0) {
	    uprintf("uiomove failed!\n");
	}
	break;
    default:
	err = ENODEV;
	break;
    }

    return err;
}

    /* To aquire data from the ADC, initiate a single conversion and
read the resulting value from the ADC FIFO buffer after the conversion
is complete.  You can generate a single conversion in three different
ways,

       1. Apply an active pulse to the CONVERT pin of the I/O connector
       2. Generate a falling edge on the sample-interval counter of the DAQ-STC
       3. Strobe the appropriate bit in a register from the E-series register set.  */


/*
 * ni_write takes in a character string and saves it
 * to buf for later accessing.
 *
 * What's the difference between dev_t and device_t?  How do I get a
 * softc struct from a dev?  
 */

int
ni_write(dev_t kdev, struct uio *uio, int ioflag)
{
    struct ni_softc *sc;
    device_t dev = devclass_get_device(ni_devclass, minor(kdev) / 16);
    int err = 0;
    unsigned char data;

    sc = (struct ni_softc *)device_get_softc(dev);

    if (!(sc->dio_flags & FWRITE)) {
	uprintf("Sorry, only digital I/O writing is supported currently!\n");
	return ENXIO;
    }

    /* Copy the string in from user memory to kernel memory */

    if ((err = uiomove(&data,1,uio)) < 0 )
	return -1;

    DAQ_STC_Windowed_Mode_Write(sc, NI_DIO_OUTPUT, data);
    
    return(err);
}

int
ni_ioctl(dev_t kdev, u_long cmd, caddr_t data, int fflag, struct proc *p)
{
    struct ni_softc *sc = devclass_get_softc(ni_devclass, minor(kdev) / 16);
    struct daq_device *pcidev = (struct daq_device *)data;

    if (cmd == NICTL_DEVICE_INFO) {
	bcopy(sc->device_info, pcidev, sizeof(*sc->device_info));
	return 0;
    }

    switch(minor(kdev) % 16) {
    case DEV_AI:
	return ai_ioctl(sc, cmd, data, fflag, p); 
    case DEV_DIO:
	return dio_ioctl(sc, cmd, data, fflag, p);
    case DEV_GPCT0:
	return gpct0_ioctl(sc, cmd, data, fflag, p);
    }
    return -EINVAL;
}

/* PCI Support Functions */

/*
 * Return identification string if this device is ours.
 */
static int
ni_probe(device_t dev)
{
    int err = ENXIO;
    uint16_t device_id;
    struct daq_device *pcicard;
    if (pci_get_vendor(dev) == NI_VENDOR_ID) {
	device_id = pci_get_device(dev);
	/* Iterate through the devices we know about to display the
	   relevant string.  (replaces 60 line switch statement) */
	for (pcicard = pci_boards; pcicard->device_id; pcicard++) {
	    if (pcicard->device_id == device_id) {
		device_set_desc(dev, pcicard->description);
		err = 0;
		break;
	    }
	}
    }

    return err;
}

/* Attach function is only called if the probe is successful */

static int
ni_attach(device_t dev)
{
    int error = 0;
    struct daq_device *pcicard;
    uint16_t device_id;
    struct ni_softc *sc = device_get_softc(dev);

    sc->unit = device_get_unit(dev);
     /* Allocate the memory resource */
    sc->bar0id = 0x10;
    sc->bar0res = bus_alloc_resource(dev, SYS_RES_MEMORY, &(sc->bar0id),
				  0, ~0, 1, RF_ACTIVE);
    if (sc->bar0res == NULL) {
	uprintf("Memory allocation of PCI base register 0 failed!\n");
	error = ENXIO;
	goto fail1;
    }

    sc->bar1id = 0x14;
    sc->bar1res = bus_alloc_resource(dev, SYS_RES_MEMORY, &(sc->bar1id),
				  0, ~0, 1, RF_ACTIVE);
    if (sc->bar1res == NULL) {
	uprintf("Memory allocation of PCI base register 1 failed!\n");
	error =  ENXIO;
	goto fail2;
    }

    /* Get the IRQ resource */

    sc->irqid = 0x0;
    sc->irqres = bus_alloc_resource(dev, SYS_RES_IRQ, &(sc->irqid),
				  0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);
    if (sc->irqres == NULL) {
	uprintf("IRQ allocation failed!\n");
	error = ENXIO;
	goto fail3;
    }

    /* Now we should setup the interrupt handler */

    error = bus_setup_intr(dev, sc->irqres, INTR_TYPE_MISC,
			   ni_handler, sc, &(sc->handler));
    if (error) {
	printf("Couldn't set up irq\n");
	goto fail4;
    }

    sc->bar0_bt = rman_get_bustag(sc->bar0res);
    sc->bar0_bh = rman_get_bushandle(sc->bar0res);
    sc->bar1_bt = rman_get_bustag(sc->bar1res);
    sc->bar1_bh = rman_get_bushandle(sc->bar1res);
    sc->irq_bt = rman_get_bustag(sc->irqres);
    sc->irq_bh = rman_get_bushandle(sc->irqres);

    /* Double check that we actually got the correct BAR values, they
       should both be 4k in size.  Take this out eventually */
    if (rman_get_size(sc->bar0res) != 4096) {
	uprintf("BAR0 should be 4k but isn't!\n");
	error = ENXIO;
	goto fail4;
    }
    if (rman_get_size(sc->bar1res) != 4096) {
	uprintf("BAR1 should be 4k but isn't!\n");
	error = ENXIO;
	goto fail4;
    }

    /* Chapter 4 of the PCI E Series Register-Level Programming Manual
       states that we must perform this step to enable the device
       window of the MITE. */

    bus_space_write_4(sc->bar0_bt, sc->bar0_bh, 0xc0, ((0xffffff00 & rman_get_start(sc->bar1res)) | 0x00000080));

    /* Setup interrupt channel 'A' (analog input) */
/*    MSC_IRQ_Configure(sc); */

    /* Make the devices */

    /* ADC Device */
    sc->daqdev = make_dev(&ni_cdevsw, (sc->unit * 16), UID_ROOT, GID_WHEEL, 
			  0600, "nidaq%d", sc->unit);
    /* Digital I/O device */
    sc->diodev = make_dev(&ni_cdevsw, (sc->unit * 16) + 1, UID_ROOT, 
			  GID_WHEEL, 0600, "nidio%d", sc->unit);

    /* General Purpose Timers */
    sc->cntdev0 = make_dev(&ni_cdevsw, (sc->unit * 16) + 2, UID_ROOT, 
			  GID_WHEEL, 0600, "nigpct%d0", sc->unit);
    sc->cntdev1 = make_dev(&ni_cdevsw, (sc->unit * 16) + 3, UID_ROOT, 
			  GID_WHEEL, 0600, "nigpct%d1", sc->unit);

    /* Statistics Device */
    sc->statdev = make_dev(&ni_cdevsw, (sc->unit * 16) + 4, UID_ROOT, 
			  GID_WHEEL, 0600, "nistats%d", sc->unit);

    /* Allocate space for the data FIFO, this cannot return NULL */
    sc->fifo = (uint16_t *)malloc(ADC_FIFO_SIZE, M_NIFIFO, M_ZERO);

    /* Record the capabilities of this device */
    device_id = pci_get_device(dev);
    for (pcicard = pci_boards; pcicard->device_id; pcicard++) {
	if (pcicard->device_id == device_id) {
	    sc->device_info = pcicard;
	    sc->ai_channels = (uint32_t *)malloc(pcicard->ai_channels * 4,
						 M_NICHAN, M_ZERO);
	    break;
	}
    }
    /* Inititalize mutex */
    mtx_init(&sc->lock, device_get_nameunit(dev), MTX_DEF);

    return error;

 fail4:
    uprintf("fail4\n");
    bus_release_resource(dev, SYS_RES_IRQ, sc->irqid, sc->irqres);
 fail3:
    uprintf("fail3\n");
    bus_release_resource(dev, SYS_RES_MEMORY, sc->bar1id, sc->bar1res);
 fail2:
    uprintf("fail2\n");
    bus_release_resource(dev, SYS_RES_MEMORY, sc->bar0id, sc->bar0res);
 fail1:
    uprintf("fail1\n");
    return error;
}

/* Detach device. */

static int
ni_detach(device_t dev)
{
    struct ni_softc *sc = device_get_softc(dev);

    /* Here we must release the memory resource */
    if (sc->bar0res) {
	bus_release_resource(dev, SYS_RES_MEMORY, sc->bar0id, sc->bar0res);
    }
    if (sc->bar1res) {
	bus_release_resource(dev, SYS_RES_MEMORY, sc->bar1id, sc->bar1res);
    }
    if (sc->irqres) {
	bus_teardown_intr(dev, sc->irqres, sc->handler);
	bus_release_resource(dev, SYS_RES_IRQ, sc->irqid, sc->irqres);
    }

    /* Free the memory we allocated */
    free(sc->fifo, M_NIFIFO);
    free(sc->ai_channels, M_NICHAN);
    Free_DMA(sc);

    /* Destroy the device nodes */
    destroy_dev(sc->daqdev);
    destroy_dev(sc->diodev);
    destroy_dev(sc->cntdev0);
    destroy_dev(sc->cntdev1);
    destroy_dev(sc->statdev);

    return 0;
}

/* Called during system shutdown after sync. */

static int
ni_shutdown(device_t dev)
{
    uprintf("Ni shutdown!\n");
    return 0;
}

static device_method_t ni_methods[] = {
    /* Device interface */
    DEVMETHOD(device_probe,		ni_probe),
    DEVMETHOD(device_attach,	ni_attach),
    DEVMETHOD(device_detach,	ni_detach),
    DEVMETHOD(device_shutdown,	ni_shutdown),
    { 0, 0 }
};

static driver_t ni_driver = {
	"ni",
	ni_methods,
	sizeof(struct ni_softc)
};

DRIVER_MODULE(ni, pci, ni_driver, ni_devclass, 0, 0);
