/*
 * Copyright (c) 2000, 2001 Murray Stokely
 * Copyright (c) 2000, 2001 BSDi
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Driver for National Instruments PCI E-Series data acquisition cards.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, May 1995 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *  [4] "MITE DMA Information", Tim Hayles October 19, 2000
 *
 */

#define NI_VENDOR_ID         0x1093

#define DEV_AI    0
#define DEV_DIO   1
#define DEV_GPCT0 2
#define DEV_GPCT1 3
#define DEV_STATS 4

/* Bits for the status register in the softc */

#define AI_BUSY      0x0001
#define DIO_BUSY     0x0002
#define GPCT0_BUSY   0x0004
#define GPCT1_BUSY   0x0008
#define AI_DMA       0x0010

/* SHOULD WE TRY DMA? */
/* #define TRYDMA */
/* More output */
#define DEBUG

#define DMA_INPUT  0
#define DMA_OUTPUT 1

/* 12bit samples are 2 bytes in size 
   XXX - should this be determined from the adc resolution */
#define SAMPLE_SIZE 2

#if defined(__alpha__)
#define vtobus(va)      alpha_XXX_dmamap((vm_offset_t)va)
#else
#define vtobus(va)      vtophys(va)
#endif

MALLOC_DECLARE(M_NIFIFO);
MALLOC_DECLARE(M_NICHAN);
MALLOC_DECLARE(M_NIDMA);
MALLOC_DECLARE(M_NIMITE);

/* We can't just use the MITELink struct to keep track of our kernel memory, 
   because we also must keep the virtual address so we can free() it! */

typedef struct DMABufferStruct {
    struct DMABufferNode *next;
    uint32_t physical;
    uint32_t virtual;
    uint32_t size;
} DMABufferNode;

/* DMA Link Chaining Mode Structure   [2] 4-59 (short link mode) 
   This structure is understood directly by the hardware.  */

typedef struct MITELinkStruct
{
    uint32_t TCR;   /* TCR : Transfer Count Register */
    uint32_t MAR;   /* MAR : Memory Address Register */
/*    uint32_t DAR; */  /* DAR : Device Address Register */
    uint32_t LKAR;  /* LKAR: Link Address Register   */
} MITELinkChain; 


struct ni_softc {
    struct mtx lock;
    struct resource *bar0res;	/* BAR0 points to the base address of
				   the MITE registers */
    int bar0id;
    bus_space_tag_t bar0_bt;
    bus_space_handle_t bar0_bh;
    struct resource *bar1res;	/* BAR1 points to the base address of
                                   the board registers such as the
                                   DAQ-STC */
    int bar1id;
    bus_space_tag_t bar1_bt;
    bus_space_handle_t bar1_bh;

    struct resource *irqres;
    int irqid;
    bus_space_tag_t irq_bt;
    bus_space_handle_t irq_bh;
    void *handler;

    int unit;
    struct daq_device *device_info;
    
    int use;

    /* Digital I/O stuff */
    int dio_flags;
    int wrmask;
    int dio;

    /* Analog input stuff */
    uint32_t scan_count;		/* 24bit scan counter  */
    uint32_t scan_interval;		/* 24bit scan interval counter */
    uint16_t sample_interval;	/* 16bit sample interval counter */
    int ai_chancnt;
    uint32_t *ai_channels;

    /* DMA stuff */
    DMABufferNode *dma_buffer_head;
    MITELinkChain *MITELinkChainHead;

    /* Device nodes from make_dev */
    dev_t daqdev;
    dev_t diodev;
    dev_t cntdev0;
    dev_t cntdev1;
    dev_t statdev;
    uint16_t *fifo;
    uint16_t fifocnt;
    uint16_t startcnt;
    int reg[256];
};

/* Prototypes */

void DAQ_STC_Windowed_Mode_Write1(struct ni_softc *sc, uint16_t address, uint8_t value);
void DAQ_STC_Windowed_Mode_Write(struct ni_softc *sc, uint16_t address, uint16_t value);
void DAQ_STC_Windowed_Mode_Masked_Write(struct ni_softc *sc, uint16_t address, uint16_t value, uint16_t mask);
void DAQ_STC_Windowed_Mode_Strobe_Write(struct ni_softc *sc, uint16_t address, uint16_t value);
uint8_t DAQ_STC_Windowed_Mode_Read1(struct ni_softc *sc, uint16_t address);
uint16_t DAQ_STC_Windowed_Mode_Read(struct ni_softc *sc, uint16_t address);
void board_write(struct ni_softc *sc, uint16_t address, uint16_t value);
void board_write1(struct ni_softc *sc, uint16_t address, uint8_t value);
uint16_t board_read(struct ni_softc *sc, uint16_t address);
uint16_t board_read1(struct ni_softc *sc, uint8_t address);
void AI_Reset_All(struct ni_softc *sc);
void AI_Ack_All(struct ni_softc *sc);
void AI_Interrupt_Enable(struct ni_softc *sc);
void Configure_Board(struct ni_softc *sc, uint32_t channel);
void MSC_Clock_Configure(struct ni_softc *sc);
void MSC_IRQ_Configure(struct ni_softc *sc);
void Clear_FIFO(struct ni_softc *sc);
void Clear_Configuration_Memory(struct ni_softc *sc);
void AI_Board_Personalize(struct ni_softc *sc);
void AI_Initialize_Configuration_Memory_Output(struct ni_softc *sc);
void AI_Board_Environmentalize(struct ni_softc *sc);
void AI_Trigger_Signals(struct ni_softc *sc);
void Number_of_Scans(struct ni_softc *sc);
void AI_Scan_Start(struct ni_softc *sc);
void AI_End_of_Scan(struct ni_softc *sc);
void AI_Start_The_Acquisition(struct ni_softc *sc);
int dio_ioctl(struct ni_softc *sc, u_long cmd, caddr_t data, int fflag, struct proc *p);
int ai_ioctl(struct ni_softc *sc, u_long cmd, caddr_t data, int fflag, struct proc *p);
void AI_Add_Channel(struct ni_softc *sc, uint32_t arg);
void AI_Arming(struct ni_softc *sc);
void Convert_Signal(struct ni_softc *sc);
void ni_handler(void *xsc);
uint32_t Encode_Channel(struct ai_channel *chan);

int gpct0_ioctl(struct ni_softc *sc, u_long cmd, caddr_t data, int fflag, struct proc *p);
void G0_Reset_All(struct ni_softc *sc);
void G0_Arm_All(struct ni_softc *sc);
void G0_Simple_Event_Counting(struct ni_softc *sc);
void G0_Buffered_Event_Counting(struct ni_softc *sc);
void G0_Single_Pulse_Generation(struct ni_softc *sc);

/* MITE Functions */
void mite_write2(struct ni_softc *sc, uint16_t address, uint16_t value);
void mite_write4(struct ni_softc *sc, uint16_t address, uint32_t value);
uint16_t mite_read2(struct ni_softc *sc, uint16_t address);
uint32_t mite_read4(struct ni_softc *sc, uint16_t address);

void mMITE_DMA_Program(struct ni_softc *sc, short, short, short);
void mMITE_DMA_Arm(struct ni_softc *sc, int channel);
void mMITE_DMA_Disarm(struct ni_softc *sc, int channel);
void mMITE_Free_Chain(struct ni_softc *sc);
void Free_DMA(struct ni_softc *sc);
void mMITE_Status(struct ni_softc *sc, int dmaChannel);
