/*
 * Copyright (c) 2000 Murray Stokely
 * Copyright (c) 2000 BSDi
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software withough specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Driver for National Instruments PCI E-Series data aquisition cards.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, January 1999 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *
 * Murray Stokely
 * 20 Feb 2001
 */

/*
 * add_drv -i 'pci1093,2a60' ni
 */

#include <sys/types.h>
#include <sys/conf.h>
#include <sys/devops.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/uio.h>
#include <sys/cred.h>
#include <sys/modctl.h>
#include <sys/cmn_err.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/stat.h>
#include <sys/uio.h>

#include "ni_solaris.h"
#include "nireg.h"
#include "niio_solaris.h"

/* Function prototypes */

int ni_open(dev_t *devp, int flag, int otyp, cred_t *cred_p);
int ni_close(dev_t devp, int flag, int otyp, cred_t *cred_p);
int ni_read(dev_t dev, struct uio *uio_p, cred_t *cred_p);
int ni_write(dev_t dev, struct uio *uio_p, cred_t *cred_p);
int ni_ioctl(dev_t dev, int cmd, intptr_t arg, int mode, cred_t *cred_p, int *rval_p);
int ni_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg, void **resultp);
int ni_probe(dev_info_t *dip);
int ni_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
int ni_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

static struct ddi_device_acc_attr endian_attr = {
    DDI_DEVICE_ATTR_V0,
    DDI_STRUCTURE_LE_ACC,
    DDI_STRICTORDER_ACC
};

static  void *ni_softstate;    /* opaque handle top of state structs */

/* Character device entry points */

static struct cb_ops ni_cb_ops = {
    ni_open,
    ni_close,
    nodev,
    nodev,
    nodev,
    ni_read,
    ni_write,
    ni_ioctl,
    nodev,
    nodev,
    nodev,
    nochpoll,
    nodev,
    NULL,
    D_MP | D_64BIT
};

static struct dev_ops ni_drv_ops = {
    DEVO_REV,
    0,
    ni_getinfo,
    nulldev,
    nulldev,
/*    ni_probe,*/
    ni_attach,
    ni_detach,
    nodev,
    &ni_cb_ops,
    NULL,
    NULL
};

/* 
 * Module linkage information for the kernel
 */

static struct modldrv ni_modldrv = {
    &mod_driverops, /* type of module.  this is a driver */
    "ni",
    &ni_drv_ops /* driver ops */
};

static struct modlinkage ni_modlinkage = {
    MODREV_1,
    &ni_modldrv,
    NULL
};


/* Must return the result of mod_install(9f) */
int 
_init() {
    int     error;
    /*
     * initalize the drivers soft state list
     * indicate that a single instance of this device is expected
     */
    if ((error = ddi_soft_state_init(&ni_softstate, sizeof (ni_softc), 1)) != 0)
	return (error);
    /*
     * install the driver linkage into the system
     */
    if ((error = mod_install(&ni_modlinkage)) != 0)
	ddi_soft_state_fini(&ni_softstate);
    return (error);
}

/* Must return the result of mod_info(9f) */
int
_info(struct modinfo *modinfop) {
    int i;
    return(mod_info(&ni_modlinkage, modinfop));
}

/* Must return the result of mod_remove(9F) */
int
_fini() {
    VERBOSE("NI module _fini!\n");
    return(mod_remove(&ni_modlinkage));
}

int
ni_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg, void **result)
{
    int     error;
    ni_softc     *ni_p;
    int     instance = getminor((dev_t)arg); 

    VERBOSE("NI getinfo!\n");
    switch (cmd) {
    case DDI_INFO_DEVT2DEVINFO:
	ni_p = ddi_get_soft_state(ni_softstate, instance);
	if (ni_p == NULL) {
	    *result = NULL;
	    error = DDI_FAILURE;
	} else {
	    mutex_enter(&ni_p->mutex);
	    *result = ni_p->dip;
	    mutex_exit(&ni_p->mutex);
	    error = DDI_SUCCESS;
	}
	break;
    case DDI_INFO_DEVT2INSTANCE:
	*result = (void *)instance;
	error = DDI_SUCCESS;
	break;
    default:
	*result = NULL;
	error = DDI_FAILURE;
    }
    return (error);
}

/*
static int 
ni_probe(dev_info_t *dip)
{
    cmn_err(CE_NOTE,"Ni probe!\n");
    return DDI_PROBE_SUCCESS;
}
*/

int 
ni_attach(dev_info_t *dip, ddi_attach_cmd_t cmd) 
{
    int instance, error;
    int n;
    caddr_t                 mem;
    ddi_acc_handle_t        handle;
    ushort_t vendor_id, device_id;
    ni_softc *ni_p;

    VERBOSE("NI attach!\n");
    switch (cmd) {
    case DDI_ATTACH:
	/*
	 * Allocate a soft state structure for this device instance
	 */
	instance = ddi_get_instance(dip);
	/* XXX Can't get macros working with multi-length arguments */
	cmn_err(CE_NOTE, "Instance : %d\n", instance);

	if (ddi_soft_state_zalloc(ni_softstate, instance) != 0) {
	    cmn_err(CE_WARN, "Failed to alloc softc!\n");
	    return (DDI_FAILURE);
	}
	ni_p = ddi_get_soft_state(ni_softstate, instance);
	if (ni_p == NULL) {
	    cmn_err(CE_WARN, "Failed to get softc!\n");
	    return (DDI_FAILURE);
	}

	ddi_set_driver_private(dip, (caddr_t)ni_p);
	ni_p->dip = dip;
	
	/* Setup the mutex */
	/* XXX The fourth argument needs to change if we ever use this
           in an interrupt handler.  See mutex_init(9f) */
	mutex_init(&ni_p->mutex, NULL, MUTEX_DRIVER, NULL);

	/* Setup the PCI Base Address Registers */
	ddi_dev_nregs(dip, &n);
	cmn_err(CE_NOTE, "Number of regs : %d\n", n);

	if (ddi_regs_map_setup(dip, 0, &mem, 0, 0, &endian_attr, &handle) != DDI_SUCCESS) {
	    cmn_err(CE_NOTE, "Could not map first base register!\n");
	    return (DDI_FAILURE);
	}
	vendor_id = ddi_get16(handle, (unsigned short *) mem + 0);
	device_id = ddi_get16(handle, (unsigned short *) (mem + 2));
	cmn_err(CE_NOTE, "Vendor ID : 0x%04x (0x%04x)\n", vendor_id, NI_VENDOR_ID);
	cmn_err(CE_NOTE, "Device ID : 0x%04x (0x%04x)\n", device_id, DEVICE_ID_6023);
	
	if (ddi_dev_regsize(dip, 1, &ni_p->mem_size1) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Invalid register number!\n");
	    return (DDI_FAILURE);
	}

	if (ddi_regs_map_setup(dip, 1, &ni_p->mem1, 0, ni_p->mem_size1, &endian_attr, &ni_p->handle1) != DDI_SUCCESS) {
	    cmn_err(CE_WARN, "Could not map first base register!\n");
	    return (DDI_FAILURE);
	}

	cmn_err(CE_NOTE, "Base Address#1 : 0x%x (size : %ld)handle: %p\n", ni_p->mem1, ni_p->mem_size1, ni_p->handle1);

	if (ddi_dev_regsize(dip, 2, &ni_p->mem_size2) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Invalid register number!\n");
	    return (DDI_FAILURE);
	}
	if (ddi_regs_map_setup(dip, 2, &ni_p->mem2, 0, ni_p->mem_size2, &endian_attr, &ni_p->handle2) != DDI_SUCCESS) {
	    cmn_err(CE_WARN, "Could not map second base register!\n");
	    return (DDI_FAILURE);
	}
	cmn_err(CE_NOTE, "Base Address#2 : 0x%x (size : %ld)handle: %p)\n", ni_p->mem2, ni_p->mem_size2, ni_p->handle2);

	/* Make the devices */
	/* ADC Device */
	if (ddi_create_minor_node(dip, "nidaq", S_IFCHR,
				  0, DDI_PSEUDO, 0) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Failed to create minor node!\n");
	    return (DDI_FAILURE);
	}

	/* Digital I/O device */
/*	if (ddi_create_minor_node(dip, ddi_get_name(dip), S_IFCHR,*/
	if (ddi_create_minor_node(dip, "nidio", S_IFCHR,
				  1, DDI_PSEUDO, 0) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Failed to create minor node!\n");
	    return (DDI_FAILURE);
	}

	/* General Purpose Timers */
	if (ddi_create_minor_node(dip, "nigpct0", S_IFCHR,
				  2, DDI_PSEUDO, 0) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Failed to create minor node!\n");
	    return (DDI_FAILURE);
	}
	if (ddi_create_minor_node(dip, "nigpct1", S_IFCHR,
				  3, DDI_PSEUDO, 0) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Failed to create minor node!\n");
	    return (DDI_FAILURE);
	}

	/* Statistics Device */
	if (ddi_create_minor_node(dip, "nistats", S_IFCHR,
				  4, DDI_PSEUDO, 0) == DDI_FAILURE) {
	    cmn_err(CE_WARN, "Failed to create minor node!\n");
	    return (DDI_FAILURE);
	}

	/*
	 * report the name of the device instance which has attached
	 */
	ddi_report_dev(dip);
	return (DDI_SUCCESS);
    case DDI_SUSPEND:
    case DDI_PM_SUSPEND:
	return (DDI_SUCCESS);
    default:
	return (DDI_FAILURE);
    }
}

int 
ni_detach(dev_info_t *dip, ddi_detach_cmd_t cmd) 
{
    int instance;
    ni_softc *ni_p;
    VERBOSE("NI detach!\n");
    switch (cmd) {
    case DDI_DETACH:
	instance = ddi_get_instance(dip);

	ni_p = ddi_get_soft_state(ni_softstate, instance);
	if (ni_p == NULL) {
	    cmn_err(CE_WARN, "Failed to get softc!\n");
	    return (DDI_FAILURE);
	}

	/*
	 * turn off the device and free any resources allocated
	 * in the attach routine
	 * XXX - do error checking!
	 */

	/* Remove the data access handles */
	ddi_regs_map_free(&ni_p->handle1);
	ddi_regs_map_free(&ni_p->handle2);

	/* Remove all the minor nodes! */
	ddi_remove_minor_node(dip, NULL);
	mutex_destroy(&ni_p->mutex);
	ddi_soft_state_free(ni_softstate, instance);
	/* FALLTHRU */
    case DDI_SUSPEND:
    case DDI_PM_SUSPEND:
	return (DDI_SUCCESS);
    default:
	return (DDI_FAILURE);
    }
}

int
ni_open(dev_t *devp, int flag, int otyp, cred_t *cred_p)
{
    int     retval = 0;
    ni_softc *sc;
    cmn_err(CE_NOTE, "in NI open!\n");

    /*
     * Verify the instance has been attached
     */
    sc = ddi_get_soft_state(ni_softstate, getminor(*devp) / 16);
    if (sc == NULL) {
	cmn_err(CE_WARN,"NI softc == NULL!\n");
	return (ENXIO);
    }

    /*
     * Verify we are being opened as a character device
     */
    if (otyp != OTYP_CHR) {
	cmn_err(CE_NOTE,"NI not opened chartype!\n");
	return (EINVAL);
    }

    /*
     * Only allow a single open
     */
    cmn_err(CE_NOTE, "NI in %d device\n", getminor(*devp) % 16);
    mutex_enter(&sc->mutex);

    switch(getminor(*devp) % 16) {
    case DEV_AI:
	if (sc->ni_state & NI_AIOPEN) {
	    cmn_err(CE_NOTE,"NI Analog Input already open!\n");
	    retval = EBUSY;
	} else {
	    cmn_err(CE_NOTE,"NI Analog Input being opened!\n");
	    sc->ni_state |= NI_AIOPEN;
	    sc->scan_count = 0;
	    sc->scan_interval = 0;
	    sc->sample_interval = 0;
	    sc->fifocnt  = 0;
	    sc->startcnt = 0;
	}
	break;
    case DEV_DIO:
	if (sc->ni_state & NI_DIOOPEN) {
	    cmn_err(CE_NOTE,"NI Digital I/O already open!\n");
	    retval = EBUSY;
	} else {
	    cmn_err(CE_NOTE,"NI Digital I/O being opened!\n");
	    sc->ni_state |= NI_DIOOPEN;
/*	    sc->dio_flags = oflags;*/
	}
	break;
    case DEV_GPCT0:
	break;
    case DEV_GPCT1:
	break;
    case DEV_STATS:
	/* A read-only device */
	cmn_err(CE_NOTE, "NI Statistics Open!\n");
	break;
    default:
	retval = ENODEV;
    }
    mutex_exit(&sc->mutex);
    return (retval);
}

int
ni_close(dev_t devp, int flag, int otyp, cred_t *cred_p)
{
    ni_softc *sc;
    int err = 0;
    sc = ddi_get_soft_state(ni_softstate, getminor(devp) / 16);

    if (sc == NULL) {
	cmn_err(CE_WARN,"(close) NI softc == NULL! (%d %d)\n", getminor(devp), getminor(devp) /16);
	return (ENXIO);
    }

    mutex_enter(&sc->mutex);

    switch(getminor(devp) % 16) {
    case DEV_AI:
	cmn_err(CE_NOTE,"NI Analog Input close!\n");
	if (sc->ni_state & NI_AIOPEN)
	    sc->ni_state -= NI_AIOPEN;
	break;
    case DEV_DIO:
	cmn_err(CE_NOTE,"NI Digital I/O close!\n");
	if (sc->ni_state & NI_DIOOPEN)
	    sc->ni_state -= NI_DIOOPEN;
	break;
    case DEV_GPCT0:
    case DEV_GPCT1:
	break;
    case DEV_STATS:
	break;
    default:
	uprintf("Unknown device!\n");
	err = ENODEV;
    }
	
    mutex_exit(&sc->mutex);    
    return err;
}

int
ni_read(dev_t dev, struct uio *uio_p, cred_t *cred_p)
{
    ni_softc *sc;
    int err = 0;
    sc = ddi_get_soft_state(ni_softstate, getminor(dev) / 16);

    if (sc == NULL) {
	cmn_err(CE_WARN,"(read) NI softc == NULL! (%d %d)\n", getminor(dev), getminor(dev) /16);
	return (ENXIO);
    }

    mutex_enter(&sc->mutex);

    switch(getminor(dev) % 16) {
    case DEV_AI:
	break;
    case DEV_DIO:
	sc->dio = DAQ_STC_Windowed_Mode_Read(sc, NI_DIO_PARALLEL_INPUT);
	cmn_err(CE_NOTE, "(dio read) We read in : 0x%02x\n", sc->dio);
	if ((err = uiomove((caddr_t) &sc->dio, 1, UIO_READ, uio_p)) != 0) {
	    uprintf("uiomove failed!\n");
	}
	cmn_err(CE_NOTE, "uiomove returned : %d\n", err);
	break;
    default:
	err = ENODEV;
	break;
    }
    mutex_exit(&sc->mutex);
    return err;
}
/*
    cmn_err(CE_NOTE,"Ni module read!\n");
    cmn_err(CE_NOTE,"uio_p->uio_iovcnt :  %d\n", uio_p->uio_iovcnt);
    cmn_err(CE_NOTE,"uio_p->uio_offset :  %d\n", uio_p->uio_offset);
    cmn_err(CE_NOTE,"uio_p->uio_loffset :  %d\n", uio_p->uio_loffset);
    cmn_err(CE_NOTE,"uio_p->uio_iov->iov_len : %d\n", uio_p->uio_iov->iov_len);
    ni_p = ddi_get_soft_state(ni_softstate, 0);
    amt = MIN(uio_p->uio_resid, (ni_p->len - uio_p->uio_offset > 0) ? ni_p->len - uio_p->uio_offset : 0);
    error = uiomove(ni_p->msg + uio_p->uio_offset, amt, UIO_READ, uio_p); 
    if (error == 0)
	cmn_err(CE_NOTE,"msg : %s\n", ni_p->msg);
*/

int
ni_write(dev_t dev, struct uio *uio_p, cred_t *cred_p)
{
    ni_softc *sc;
    int err = 0;
    unsigned char data;
    sc = ddi_get_soft_state(ni_softstate, getminor(dev) / 16);

    cmn_err(CE_NOTE,"Ni module write!\n");
    if (sc == NULL) {
	cmn_err(CE_WARN,"(close) NI softc == NULL! (%d %d)\n", getminor(dev), getminor(dev) /16);
	return (ENXIO);
    }

    mutex_enter(&sc->mutex);

    switch(getminor(dev) % 16) {
    case DEV_DIO:
	if ((err = uiomove(&data, 1, UIO_WRITE, uio_p)) < 0 ) {
	    cmn_err(CE_WARN, "uiomove failed!\n");
	    return -1;
	}
	DAQ_STC_Windowed_Mode_Write(sc, NI_DIO_OUTPUT, data);
	break;
    default:
	cmn_err(CE_WARN, "Only Digital Output is supported!\n");
	err = ENODEV;
	break;
    }
    mutex_exit(&sc->mutex);
    return err;
}

/*
    ni_p = ddi_get_soft_state(ni_softstate, 0); 
    error = uiomove(ni_p->msg, MAX_SIZE, UIO_WRITE, uio_p);
    cmn_err(CE_NOTE,"uio_p->uio_iovcnt :  %d\n", uio_p->uio_iovcnt);
    cmn_err(CE_NOTE,"uio_p->uio_offset :  %d\n", uio_p->uio_offset);
    cmn_err(CE_NOTE,"uio_p->uio_loffset :  %d\n", uio_p->uio_loffset);
    cmn_err(CE_NOTE,"uio_p->uio_iov->iov_len : %d\n", uio_p->uio_iov->iov_len);
    *(ni_p->msg + uio_p->uio_offset) = 0;
    ni_p->len = uio_p->uio_offset;

    if (error == 0)
	cmn_err(CE_NOTE,"msg : %s\n", ni_p->msg);
    cmn_err(CE_NOTE,"error : %d\n", error);
*/

int
ni_ioctl(dev_t dev, int cmd, intptr_t arg, int mode, cred_t *cred_p, int *rval_p)
{
    ni_softc *sc;
    int err = 0;
    unsigned char data;
    sc = ddi_get_soft_state(ni_softstate, getminor(dev) / 16);

    cmn_err(CE_NOTE,"NI module IOCTL (%d)!\n", cmd);
    if (sc == NULL) {
	cmn_err(CE_WARN,"(close) NI softc == NULL! (%d %d)\n", getminor(dev), getminor(dev) /16);
	return (ENXIO);
    }

    switch(cmd) {
    case NICTL_DIO_WRMASK:
	/* No need for a ddi_copyin, the user passes a value in
           directly as opposed to an address (in userspace) */
	cmn_err(CE_NOTE, "DIO IOCTL WRMASK!\n");
	sc->wrmask = arg;
	mutex_enter(&sc->mutex);
	DAQ_STC_Windowed_Mode_Write(sc, NI_DIO_CONTROL, sc->wrmask);
	mutex_exit(&sc->mutex);
	break;
    default:
	cmn_err(CE_NOTE, "Invalid IOCTL!\n");
	err = EINVAL;
    }
    return err;
}
