/*
 * Copyright (c) 2001 Murray Stokely
 * Copyright (c) 2001 BSDi
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in this position and unchanged.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Driver for National Instruments PCI E-Series data acquisition cards.
 *
 * References :
 *  [1] DAQ-STC Technical Reference Manual, May 1995 Edition
 *  [2] PCI E Series Register-Level Programmer Manual, November 1998 Edition
 *  [3] PCI E Series User Manual, January 1999 Edition
 *
 */

/*
 * To determine if a PCI E Series board has a pending interrupt, we should perform
 * a 32bit read from BAR0 + 0x14 and then check the status of bit 31.
 * [2] 4-57
 */
#define NIB_INTERRUPT                    0x80000000

/* -----------------------------------------------------------------------
 * PCI E Series Registers 
 * From : [2] Chapter 3
 *
 * These registers can be accessed directly from BAR1
 * -----------------------------------------------------------------------
 */

/* Misc Register Group */
#define NI_SERIAL_COMMAND                0x0D	/* 8bit, write-only */
#define NI_MISC_COMMAND                  0x0F	/* 8bit, write-only */
#define NI_STATUS                        0x01	/* 8bit, read-only */

/* Analog Input Register Group */
#define NI_ADC_FIFO_DATA                 0x1C	/* 16bit, read-only */
#define NI_CONFIGURATION_MEMORY_LOW      0x10	/* 16bit, write-only */
#define NI_CONFIGURATION_MEMORY_HIGH     0x12	/* 16bit, write-only */

/* Bitfields for NI_CONFIGURATION_MEMORY_LOW */
#define NIB_LAST_CHANNEL                 0x8000
#define NIB_DITHER_ENABLE                0x0200
#define NIB_POLARITY                     0x0100
#define NIB_GAIN1                        0x0001

/* Bitfields for NI_CONFIGURATION_MEMORY_HI */
#define NIB_CHANTYPE_MASK                0x7000
#define NIB_CALLIBRATION                 0x0000
#define NIB_DIFFERENTIAL                 0x1000
#define NIB_NRSE                         0x2000
#define NIB_RSE                          0x3000
#define NIB_X                            0x4000
#define NIB_AUX                          0x5000
#define NIB_GHOST                        0x7000
#define NIB_BANK1                        0x0020
#define NIB_BANK0                        0x0010
#define NIB_CHAN3                        0x0008
#define NIB_CHAN2                        0x0004
#define NIB_CHAN1                        0x0002
#define NIB_CHAN0                        0x0001

/* -----------------------------------------------------------------------
 * DAQ-STC Registers
 * From : [1] Chapter 3
 *
 * These registers must be accessed through the window registers of BAR1
 * -----------------------------------------------------------------------
 */

/* Write Registers */

    
#define NI_WINDOW_ADDRESS                0x00	/* 16bit, read/write */
#define NI_WINDOW_DATA                   0x01	/* 16bit, read/write */
#define NI_INTERRUPT_A_ACK               0x02	/* 16bit, write */
#define NI_INTERRUPT_B_ACK               0x03	/* 16bit, write */
#define NI_AI_COMMAND2                   0x04	/* 16bit, write */
#define NI_AO_COMMAND2                   0x05	/* 16bit, write */
#define NI_G0_COMMAND                    0x06	/* 16bit, write */
#define NI_G1_COMMAND                    0x07	/* 16bit, write */
#define NI_AI_COMMAND1                   0x08   /* 16bit, write */
#define NI_AO_COMMAND1                   0x09   /* 16bit, write */
#define NI_DIO_OUTPUT                    0x0A   /* 16bit, write */
#define NI_DIO_CONTROL                   0x0B   /* 16bit, write */
#define NI_AI_MODE1                      0x0C   /* 16bit, write */
#define NI_AI_MODE2                      0x0D   /* 16bit, write */
#define NI_AI_SI_LOAD_A1                 14
#define NI_AI_SI_LOAD_A2                 15
#define NI_AI_SC_LOAD_A1                 18
#define NI_AI_SC_LOAD_A2                 19     /* 24 bit field */
#define NI_AI_SI2_LOAD_A                 23
#define NI_AI_SI2_LOAD_B                 25
#define NI_CLOCK_AND_FOUT                56
#define NI_INTERRUPT_CONTROL             59
#define NI_AI_OUTPUT_CONTROL             60 /* 16bit write-only */
#define NI_AI_START_STOP_SELECT          62 /* 16bit write-only */
#define NI_AI_TRIGGER_SELECT             63 /* 16bit write-only */
#define NI_JOINT_RESET                   72 /* 16bit write-only */
#define NI_INTERRUPT_A_ENABLE            73 /* 16bit write-only */
#define NI_INTERRUPT_B_ENABLE            75 /* 16bit write-only */
#define NI_SECOND_IRQ_B_ENABLE           76
#define NI_AI_PERSONAL                   77 /* 16bit write-only */
#define NI_WRITE_STROBE0                 82 /* 16bit write-only */
#define NI_WRITE_STROBE1                 83 /* 16bit write-only */
#define NI_AI_MODE3                      87 /* 16bit write-only */

/* Read Registers */

#define NI_AI_STATUS1                    2
#define NI_AI_STATUS2                    5
#define NI_DIO_PARALLEL_INPUT            7
#define NI_JOINT_STATUS1                 27
#define NI_DIO_SERIAL_INPUT              28
#define NI_JOINT_STATUS2                 29

/* Bitfields for NI_INTERRUPT_A_ACK */
#define NIB_G0_GATE_ERROR_CONFIRM      0x0020
#define NIB_G0_TC_ERROR_CONFIRM        0x0040
#define NIB_AI_SC_TC_ERROR_CONFIRM     0x0080
#define NIB_AI_SC_TC_INTERRUPT_ACK     0x0100
#define NIB_AI_START1_INTERRUPT_ACK    0x0200
#define NIB_AI_START2_INTERRUPT_ACK    0x0400
#define NIB_AI_START_INTERRUPT_ACK     0x0800
#define NIB_AI_STOP_INTERRUPT_ACK      0x1000
#define NIB_AI_ERROR_INTERRUPT_ACK     0x2000
#define NIB_G0_GATE_INTERRUPT_ACK      0x4000
#define NIB_G0_TC_INTERRUPT_ACK        0x8000

/* Bitfields for NI_INTERRUPT_CONTROL */
#define NIB_INTERRUPT_A_ENABLE         0x0800
#define NIB_INTERRUPT_OUTPUT_POLARITY  0x0001

/* Bitfields for NI_INTERRUPT_A_ENABLE */
#define NIB_AI_SC_TC_INTERRUPT_ENABLE  0x0001
#define NIB_AI_START1_INTERRUPT_ENABLE 0x0002
#define NIB_AI_START2_INTERRUPT_ENABLE 0x0004
#define NIB_AI_START_INTERRUPT_ENABLE  0x0008
#define NIB_AI_STOP_INTERRUPT_ENABLE   0x0010
#define NIB_AI_ERROR_INTERRUPT_ENABLE  0x0020
#define NIB_G0_TC_INTERRUPT_ENABLE     0x0040
#define NIB_AI_FIFO_INTERRUPT_ENABLE   0x0080
#define NIB_G0_GATE_INTERRUPT_ENABLE   0x0100

/* Bitfields for NI_CLOCK_AND_FOUT */
#define NIB_AI_OUTPUT_DIVIDE             0x0080
#define NIB_CLOCK_TO_BOARD               0x0100
#define NIB_CLOCK_TO_BOARD_DIVIDE        0x0200
#define NIB_SLOW_INTERNAL_TIMEBASE       0x0800
#define NIB_SLOW_INTERNAL_TIME_DIVIDE    0x1000
#define NIB_FOUT_ENABLE                  0x8000

/* Bitfields for NI_JOINT_RESET */
#define NIB_AI_RESET                   0x0001
#define NIB_AI_CONFIGURATION_START     0x0010
#define NIB_AI_CONFIGURATION_END       0x0100
#define NIB_G0_RESET                   0x0004

/* Bitfields for NI_AI_MODE1 */
#define NIB_AI_TRIGGER_ONCE            0x0001
#define NIB_AI_CONTINUOUS              0x0002
#define NIB_AI_RESERVED_ONE            0x0004
#define NIB_AI_START_STOP              0x0008

/* Bitfields for NI_AI_COMMAND1 */
#define NIB_AI_CONVERT_PULSE           0x0001
#define NIB_AI_SC_LOAD                 0x0020
#define NIB_AI_SC_ARM                  0x0040
#define NIB_AI_DIV_ARM                 0x0100
#define NIB_AI_SI_LOAD                 0x0200
#define NIB_AI_SI_ARM                  0x0400
#define NIB_AI_SI2_LOAD                0x0800
#define NIB_AI_SI2_ARM                 0x1000

/* Bitfields for NI_AI_COMMAND1 */
#define NIB_AI_START1_PULSE            0x0001

/* Bitfields for NI_AI_MODE2 */
#define NIB_AI_EXTERNAL_MUX_PRESENT    0x1000
#define NIB_AI_SI2_RELOAD_MODE         0x0100
#define NIB_AI_SI2_INITIAL_LOAD_SOURCE 0x0200

/* Bitfields for NI_AI_START_STOP_SELECT */
#define NIB_AI_START_EDGE              0x0020
#define NIB_AI_START_SYNC              0x0040

/* Bitfields for NI_AI_TRIGGER_SELECT */
#define NIB_AI_START1_EDGE             0x0020
#define NIB_AI_START1_SYNC             0x0040

/* Bitfields for NI_AI_STATUS1 */
#define NIB_INTERRUPT_A_ST              0x8000
#define NIB_AI_FIFO_FULL_ST             0x4000
#define NIB_AI_FIFO_HALF_FULL_ST        0x2000
#define NIB_AI_FIFO_EMPTY_ST            0x1000
#define NIB_AI_OVERRUN_ST               0x0800
#define NIB_AI_OVERFLOW_ST              0x0400
#define NIB_AI_START_ST                 0x0020
#define NIB_AI_STOP_ST                  0x0010
#define NIB_G0_TC_ST                    0x0008
#define NIB_G0_GATE_INTERRUPT_ST        0x0004
#define NIB_AI_FIFO_REQUEST_ST          0x0002
#define NIB_PASS_THROUGH_0_INTERRUPT_ST 0x0001

/* GPCT Registers */
#define NI_G0_AUTOINCREMENT            0x0068
#define NI_G0_HW_SAVE1                 0x0008
#define NI_G0_HW_SAVE2                 0x0009
#define NI_G0_INPUT_SELECT             0x0036
#define NI_G0_LOAD_A1                  0x0029
#define NI_G0_LOAD_A2                  0x0028
#define NI_G0_LOAD_B1                  0x0031
#define NI_G0_LOAD_B2                  0x0030
#define NI_G0_MODE                     0x0026
#define NI_G0_SAVE1                    0x0013
#define NI_G0_SAVE2                    0x0012

/* Bitfields for G0_MODE */
#define NIB_G0_GATING_MODE                0x0003
#define NIB_G0_GATE_ON_BOTH_EDGES         0x0004
#define NIB_G0_TRIGGER_MODE_FOR_EDGE_GATE 0x0018
#define NIB_G0_STOP_MODE                  0x0060
#define NIB_G0_LOAD_SOURCE_SELECT         0x0080
#define NIB_G0_OUTPUT_MODE                0x0300
#define NIB_G0_COUNTING_ONCE              0x0C00
#define NIB_G0_LOADING_ON_TC              0x1000
#define NIB_G0_GATE_POLARITY              0x2000
#define NIB_G0_LOADING_ON_GATE            0x4000
#define NIB_G0_RELOAD_SOURCE_SWITCHING    0x8000

/* Bitfields for G0_COMMAND */
#define NIB_G1_DISARM_COPY            0x8000
#define NIB_G1_SAVE_TRACE_COPY        0x4000
#define NIB_G1_ARM_COPY               0x2000
#define NIB_G0_BANK_SWITCH_ENABLE     0x1000
#define NIB_G0_BANK_SWITCH_MODE       0x0800
#define NIB_G0_BANK_SWITCH_START      0x0400
#define NIB_G0_LITTLE_BIG_ENDIAN      0x0200
#define NIB_G0_SYNCHRONIZED_GATE      0x0100
#define NIB_G0_WRITE_SWITCH           0x0080
#define NIB_G0_UP_DOWN                0x0060
#define NIB_G0_DISARM                 0x0010
#define NIB_G0_ANALOG_TRIGGER_RESET   0x0008
#define NIB_G0_LOAD                   0x0004
#define NIB_G0_SAVE_TRACE             0x0002
#define NIB_G0_ARM                    0x0001

/* Bitfields for G0_INPUT_SELECT */
#define NIB_G0_SOURCE_POLARITY         0x8000
#define NIB_G0_OUTPUT_POLARITY         0x4000
#define NIB_G0_OR_GATE                 0x2000
#define NIB_G0_GATE_SELECT_LOAD_SOURCE 0x1000
#define NIB_G0_GATE_SELECT             0x0F80
#define NIB_G0_SOURCE_SELECT           0x007C
#define NIB_G0_WRITE_ACKNOWLEDGES_IRQ  0x0002
#define NIB_G0_READ_ACKNOWLEDGES_IRQ   0x0001
